"""Component switch"""
from ..log import Log
from ..device_control import CoolkitDeviceControl
from typing import TYPE_CHECKING, Callable, Dict, Awaitable

if TYPE_CHECKING:
    from .device import CoolkitDevice


class CoolkitDeviceSwitch:
    def __init__(self, device: 'CoolkitDevice', index: int):
        self._index = index
        self._device = device
        self._state = False
        self._callbacks: Dict[str, Callable[['CoolkitDeviceSwitch', bool], Awaitable[None]]] = {}

    def get_state(self) -> bool:
        return self._state

    async def update_state(self, new_state: bool) -> None:
        if new_state != self._state:
            self._state = new_state
            for callback in self._callbacks.values():
                await callback(self, new_state)

    def add_state_callback(
            self,
            callback_name: str,
            callable: Callable[['CoolkitDeviceSwitch', bool], Awaitable[None]]
    ) -> None:
        self._callbacks[callback_name] = callable

    def remove_callback(self, callback_name: str) -> None:
        if callback_name in self._callbacks:
            del self._callbacks[callback_name]

    async def set_state(self, state: bool) -> None:
        state_name = 'on' if state else 'off'

        if len(self._device.switches) == 1:
            new_params = {'switch': state_name}
        else:
            new_params = {'switches': self._device.params['switches']}
            new_params['switches'][self._index]['switch'] = state_name

        await CoolkitDeviceControl.update_command(
            device_id=self._device.device_id,
            params=new_params
        )

        await self.update_state(state)
        Log.info('Sent ' + str(self._device) + ' state=' + state_name)



import asyncio
from threading import Thread

from aiohttp import ClientSession

from .devices_repository import CoolkitDevicesRepository
from .device import CoolkitDevice
from .log import Log
from .session import CoolkitSession


class CoolkitDevicesDiscovery:
    @classmethod
    async def discover(cls) -> bool:
        devices_endpoint = CoolkitSession.get_api_endpoint_url('api/user/device')

        async with ClientSession(headers=CoolkitSession.get_auth_headers()) as session:
            async with session.get(devices_endpoint) as response:
                data = await response.json()

                if response.status != 200 or ('error' in data and data['error'] != 0):
                    Log.error('Error while trying to retrieve devices list: ' + str(data['error']))
                    return False

                for device_data in data:
                    if not CoolkitDevicesRepository.has_device(device_data['deviceid']):
                        device = CoolkitDevice(device_data)
                        CoolkitDevicesRepository.add_device(device)
                        Log.info('Registered device: ' + str(device))

        return True

    @classmethod
    async def _discover_in_background(cls) -> None:
        while True:
            await cls.discover()
            await asyncio.sleep(60)

    @classmethod
    def _start_daemon(cls):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(cls._discover_in_background())

    @classmethod
    def start_daemon(cls):
        worker = Thread(target=cls._start_daemon)
        worker.setDaemon(True)
        worker.start()



"""Devices control"""
import asyncio
import json
import random
import time
import websockets
from asyncio import AbstractEventLoop
from .devices_repository import CoolkitDevicesRepository
from .log import Log
from .session import CoolkitSession


class CoolkitDeviceControl:
    _websocket: websockets.WebSocketClientProtocol = None
    _ready = False

    @classmethod
    def _get_auth_payload(cls) -> dict:
        return {
            'action': "userOnline",
            'userAgent': 'app',
            'version': 6,
            'nonce': ''.join([str(random.randint(0, 9)) for _ in range(15)]),
            'apkVesrion': "1.8",
            'os': 'ios',
            'at': CoolkitSession.get_bearer_token(),
            'apikey': CoolkitSession.get_user_api_key(),
            'ts': str(int(time.time())),
            'model': 'iPhone10,6',
            'romVersion': '11.1.2',
            'sequence': str(time.time()).replace('.', '')
        }

    @classmethod
    async def disconnect(cls) -> None:
        await cls._websocket.close()
        cls._websocket = None

    @classmethod
    async def connect(cls) -> bool:
        """Connect websocket for real time updates"""
        if cls._websocket is None:
            ws_endpoint = CoolkitSession.get_ws_endpoint()

            cls._websocket = await websockets.connect(ws_endpoint)
            await cls._ws_send(cls._get_auth_payload())

            data = await cls._ws_read()
            if 'error' in data and data['error'] != 0:
                Log.error('Error while trying to connect websocket: ' + str(data['error']))
                await cls.disconnect()
                return False

            Log.info('Websocket successfully connected')

        cls._ready = True
        return True

    @classmethod
    async def _ws_read(cls) -> dict:
        await cls.connect()
        return json.loads(await cls._websocket.recv())

    @classmethod
    async def _ws_send(cls, payload: dict) -> None:
        await cls.connect()
        await cls._websocket.send(json.dumps(payload))

    @classmethod
    async def _listen_ws(cls):
        loop = asyncio.get_event_loop()

        await cls.connect()
        while True:
            data = await cls._ws_read()
            if data is not None:
                if 'deviceid' in data and 'params' in data:
                    device_id = data.get('deviceid')
                    CoolkitDevicesRepository.get_device(device_id).update_params(data['params'])

                if data.get('action') == 'update':
                    if data['params'].get('switch'):
                        await loop.create_task(cls._on_update_switch(data, loop))
                    elif data['params'].get('switches'):
                        await loop.create_task(cls._on_update_switches(data, loop))

    @classmethod
    async def _on_update_switch(cls, data: dict, loop: AbstractEventLoop) -> None:
        """Update device with a single outlet"""
        device = CoolkitDevicesRepository.get_device(data['deviceid'])
        if device is None:
            return

        await loop.create_task(device.switches[0].update_state(data['params']['switch'] == 'on'))

    @classmethod
    async def _on_update_switches(cls, data: dict, loop: AbstractEventLoop) -> None:
        """Update device with multiple outlets"""
        device = CoolkitDevicesRepository.get_device(data['deviceid'])
        if device is None:
            return

        for switch in data['params']['switches']:
            index = int(switch['outlet'])
            await loop.create_task(device.switches[index].update_state(switch['switch'] == 'on'))

    @classmethod
    def start_daemon(cls):
        loop = asyncio.get_event_loop()
        loop.create_task(cls._listen_ws())

    @classmethod
    async def update_command(
        cls,
        params: dict,
        device_id: str
    ) -> None:
        if not cls._ready:
            return

        device = CoolkitDevicesRepository.get_device(device_id)

        payload = {
            'action': 'update',
            'userAgent': 'app',
            'params': params,
            'apikey': device.api_key,
            'deviceid': device.device_id,
            'sequence': str(time.time()).replace('.', ''),
            'controlType': device.params['controlType'] if 'controlType' in device.params else 4,
            'ts': 0
        }

        if device.api_key != CoolkitSession.get_user_api_key():
            payload['selfApikey'] = CoolkitSession.get_user_api_key()

        await cls._ws_send(payload)



"""Static class for COOLKIT authoentication"""

import base64
import hashlib
import hmac
import json
import random
import re
import time
import uuid
from aiohttp import ClientSession

from .log import Log
from .const import COOLKIT_APP_ID, COOLKIT_APP_SECRET


class CoolkitSession:
    _bearer_token: str = None
    _user_apikey: str = None
    _region: str = None
    _ws_host: str = None

    @classmethod
    def _get_login_data(cls, username: str, password: str) -> dict:
        """Get application details for auth"""
        login_data = {
            'password': password,
            'version': '6',
            'ts': int(time.time()),
            'nonce': ''.join([str(random.randint(0, 9)) for _ in range(15)]),
            'appid': COOLKIT_APP_ID,
            'imei': str(uuid.uuid4()),
            'os': 'iOS',
            'model': 'iPhone10,6',
            'romVersion': '11.1.2',
            'appVersion': '3.5.3'
        }

        if re.match(r'[^@]+@[^@]+\.[^@]+', username):
            login_data['email'] = username
        else:
            login_data['phoneNumber'] = username

        return login_data

    @classmethod
    def _get_login_headers(cls, login_data: dict) -> dict:
        """Get headers with authorization sign"""
        hex_dig = hmac.new(
            COOLKIT_APP_SECRET,
            str.encode(json.dumps(login_data)),
            digestmod=hashlib.sha256).digest()

        sign = base64.b64encode(hex_dig).decode()

        return {
            'Authorization': 'Sign ' + sign,
            'Content-Type': 'application/json;charset=UTF-8'
        }

    @classmethod
    def get_auth_headers(cls) -> dict:
        """Get authorization headers for subsequent calls"""
        return {
            'Authorization': 'Bearer ' + cls._bearer_token,
            'Content-Type': 'application/json;charset=UTF-8'
        }

    @classmethod
    def get_bearer_token(cls):
        return cls._bearer_token

    @classmethod
    def get_user_api_key(cls):
        return cls._user_apikey

    @classmethod
    def get_ws_endpoint(cls):
        """Get websocket endpoint"""
        return 'wss://{}:8080/api/ws'.format(cls._ws_host)

    @classmethod
    def get_api_endpoint_url(cls, action: str) -> str:
        """Get API URL depending on region"""
        return 'https://{}-api.coolkit.cc:8080/{}'.format(cls._region, action)

    @classmethod
    def get_dispatch_endpoint_url(cls, action: str) -> str:
        """Get dispatch endpoint URL depending on region"""
        return 'https://{}-disp.coolkit.cc:8080/{}'.format(cls._region, action)

    @classmethod
    async def _dispatch_application(cls) -> bool:
        dispatch_url = cls.get_dispatch_endpoint_url('dispatch/app')

        async with ClientSession(headers=cls.get_auth_headers()) as session:
            async with session.post(dispatch_url) as response:
                data = await response.json()

                if response.status != 200 or ('error' in data and data['error'] != 0):
                    Log.error('Error while trying to dispatch application: ' + str(data['error']))
                    return False

                ws_host = data['domain']
                Log.info('Application assigned to ws host ' + ws_host)

                cls._ws_host = ws_host
                return True

    @classmethod
    async def login(cls, username: str, password: str, region: str) -> bool:
        """Login to COOLKIT platform"""
        cls._region = region

        login_url = cls.get_api_endpoint_url('api/user/login')
        login_data = cls._get_login_data(
            username=username,
            password=password
        )
        login_headers = cls._get_login_headers(login_data)

        async with ClientSession(headers=login_headers) as session:
            async with session.post(login_url, json=login_data) as response:
                data = await response.json()

                if response.status != 200 or ('error' in data and data['error'] != 0):
                    Log.error('Error while trying to login: ' + str(data['error']) + ' ' + data['info'])
                    return False

                cls._bearer_token = data['at']
                cls._user_apikey = data['user']['apikey']
                Log.info('User ' + username + ' successfully logged in')

                return await cls._dispatch_application()





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="coolkit-client-phoenix",
    version="1.0.1",
    author="Riccardo Tempesta",
    author_email="info@riccardotempesta.com",
    description="Sonoff control library through coolkit/ewelink cloud",
    long_description=long_description,
    url="https://github.com/phoenix128/python-coolkit-client",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import logging


class Log:
    LOGGER_NAME: str = 'sonoff-client'

    logger: logging.Logger = None

    @classmethod
    def init_log(cls) -> None:
        if cls.logger is None:
            cls.logger = logging.getLogger(cls.LOGGER_NAME)
            cls.logger.setLevel(logging.INFO)

            fh = logging.FileHandler('sonoff_client.log')

            ch = logging.StreamHandler()
            ch.setLevel(logging.DEBUG)

            formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')

            fh.setFormatter(formatter)
            ch.setFormatter(formatter)

            logger.addHandler(fh)
            logger.addHandler(ch)

    @classmethod
    def get_logger(cls) -> logging.Logger:
        return cls.logger

    @classmethod
    def debug(cls, message: str) -> None:
        cls.logger.debug(message)

    @classmethod
    def info(cls, message: str) -> None:
        cls.logger.info(message)

    @classmethod
    def warning(cls, message: str) -> None:
        cls.logger.warning(message)

    @classmethod
    def error(cls, message: str) -> None:
        cls.logger.error(message)


logger = logging.getLogger(Log.LOGGER_NAME)
logger.setLevel(logging.DEBUG)
Log.init_log()



"""Devices object"""
from typing import List, Dict
from deepmerge import always_merger

from .switch import CoolkitDeviceSwitch


class CoolkitDevice:
    def __init__(self, payload: dict):
        self._payload = payload
        self._switches: List[CoolkitDeviceSwitch] = []
        self._populate_components()

    def get_info(self, param: str):
        if param not in self._payload:
            return None

        return self._payload[param]

    def _populate_components(self):
        if 'switch' in self.params:
            self._switches.append(CoolkitDeviceSwitch(self, 0))
        elif 'switches' in self.params:
            for i in range(0, len(self.params['switches'])):
                self._switches.append(CoolkitDeviceSwitch(self, i))

    @property
    def switches(self):
        return self._switches

    @property
    def params(self) -> dict:
        return self.get_info('params')

    def update_params(self, params: Dict):
        self._payload = always_merger.merge(self._payload, params)

    @property
    def api_key(self) -> str:
        return self.get_info('apikey')

    @property
    def device_id(self) -> str:
        return self.get_info('deviceid')

    @property
    def name(self) -> str:
        return self.get_info('name')

    @property
    def device_type(self) -> str:
        return self.get_info('type')

    @property
    def device_model(self) -> str:
        return self.get_info('extra')['extra']['model']

    @property
    def product_model(self) -> str:
        return self.get_info('productModel')

    @property
    def brand(self) -> str:
        return self.get_info('brandName')

    @property
    def is_online(self) -> bool:
        return self.get_info('online')

    def __repr__(self) -> str:
        return '[' + self.device_id + '] ' + self.brand + ' ' + self.product_model



"""Devices repository"""
from typing import Dict, Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .device import CoolkitDevice


class CoolkitDevicesRepository:
    _devices: Dict[str, 'CoolkitDevice'] = {}

    @classmethod
    def get_devices(cls) -> Dict[str, 'CoolkitDevice']:
        return cls._devices

    @classmethod
    def has_device(cls, device_id: str) -> bool:
        return device_id in cls._devices

    @classmethod
    def get_device(cls, device_id: str) -> Optional['CoolkitDevice']:
        if device_id not in cls._devices:
            return None

        return cls._devices[device_id]

    @classmethod
    def add_device(cls, device: 'CoolkitDevice') -> None:
        cls._devices[device.device_id] = device
